/***************************************************
  Adafruit MQTT Library WINC1500 Example

  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/


#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>
#include <SD.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"
#include <WiFi101.h>
#include <RTCZero.h>
#include "Adafruit_STMPE610.h"
#include <Wire.h>
#include <Adafruit_INA219.h>



/* Create an rtc object */
RTCZero rtc;


#ifdef ESP8266
#define STMPE_CS 16
#define TFT_CS   0
#define TFT_DC   15
#define SD_CS    2
#endif
#ifdef ESP32
#define STMPE_CS 32
#define TFT_CS   15
#define TFT_DC   33
#define SD_CS    14
#endif
#ifdef TEENSYDUINO
#define TFT_DC   10
#define TFT_CS   4
#define STMPE_CS 3
#define SD_CS    8
#endif
#ifdef ARDUINO_STM32_FEATHER
#define TFT_DC   PB4
#define TFT_CS   PA15
#define STMPE_CS PC7
#define SD_CS    PC5
#endif
#ifdef ARDUINO_NRF52832_FEATHER /* BSP 0.6.5 and higher! */
#define TFT_DC   11
#define TFT_CS   31
#define STMPE_CS 30
#define SD_CS    27
#endif
#if defined(ARDUINO_MAX32620FTHR) || defined(ARDUINO_MAX32630FTHR)
#define TFT_DC   P5_4
#define TFT_CS   P5_3
#define STMPE_CS P3_3
#define SD_CS    P3_2
#endif

// Anything else!
#if defined (__AVR_ATmega32U4__) || defined(ARDUINO_SAMD_FEATHER_M0) || defined (__AVR_ATmega328P__) || \
    defined(ARDUINO_SAMD_ZERO) || defined(__SAMD51__) || defined(__SAM3X8E__) || defined(ARDUINO_NRF52840_FEATHER)
#define STMPE_CS 6
#define TFT_CS   9
#define TFT_DC   10
#define SD_CS    5
#endif

// Color definitions - http://www.barth-dev.de/online/rgb565-color-picker/
#define BLACK     0x0000
#define BLUE      0x001F
#define RED       0xF800
#define GREEN     0x07E0
#define CYAN      0x07FF
#define MAGENTA   0xF81F
#define YELLOW    0xFFE0
#define WHITE     0xFFFF
#define GREY      0x8410
#define DKGREEN   0x0400

#define TS_MINX 150
#define TS_MINY 130
#define TS_MAXX 3800
#define TS_MAXY 4000


Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC);

Adafruit_INA219 ina219;

Adafruit_STMPE610 ts = Adafruit_STMPE610(STMPE_CS);

#define VBATPIN A7

/************************* WiFI Setup *****************************/
#define WINC_CS   8
#define WINC_IRQ  7
#define WINC_RST  4
#define WINC_EN   2     // or, tie EN to VCC

char ssid[] = "[SSID]";     //  your network SSID (name)
char pass[] = "[PASSWORD]";    // your network password (use for WPA, or use as key for WEP)
int keyIndex = 0;                // your network key Index number (needed only for WEP)

int status = WL_IDLE_STATUS;

/************************* Adafruit.io Setup *********************************/

#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883
#define AIO_USERNAME    "[AIO_USER]"
#define AIO_KEY         "[AIO_KEY]"

/************************* Logging Details *********************************/

boolean sdActive = true; //set to true if saving to SD
boolean aioActive = true; //set to true if saving to Adafruit IO
boolean enableTrigger = true; // set to true if logger can trigger an event
boolean trackTriggerIn = false; // set to true if logger can trigger an event
boolean isRecording = true; //plcaeholder for button functionality

/************************* Voltage Logger Configuration *********************************/

#define NORMAL_MILLISECS_BETWEEN_READING              300

#define NORMAL_MILLISECS_BETWEEN_READING_AIO          5000 //this is the normal number of millisecs between readings to AIO
#define NORMAL_MILLISECS_BETWEEN_READING_SD           2000 //this is the normal number of millisecs between readings to SD card

#define TRIGGER_EVENT_MILLISECS_BETWEEN_READING_AIO   5000 //this is the  number of millisecs between readings if an event is registered
#define TRIGGER_EVENT_MILLISECS_BETWEEN_READING_SD    100 //this is the  number of millisecs between readings if an event is registered

#define NORMAL_MILLISECS_BETWEEN_TRIGGER              60000 //this is how often we manually trigger. e.g. = trigger every X minute [60,000 = 1 min]

#define CURRENT_CALIBRATION                           .21
#define VOLTAGE_CALIBRATION                           0.011

#define datafile                                      "data.csv"
#define triggerfile                                   "trigger.csv"

float shuntvoltage = 0;
float busvoltage = 0;
float current_mA = 0;
float loadvoltage = 0;
float power_mW = 0;

boolean recievedTrigger = false; //this is changed by code


/************ Global State (you don't need to change this!) ******************/

//Set up the wifi client
WiFiClient client;

Adafruit_MQTT_Client mqtt(&client, AIO_SERVER, AIO_SERVERPORT, AIO_USERNAME, AIO_KEY);

// You don't need to change anything below this line!
#define halt(s) { Serial.println(F( s )); while(1);  }

/****************************** Feeds ***************************************/

// Notice MQTT paths for AIO follow the form: <username>/feeds/<feedname>
Adafruit_MQTT_Publish voltage = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "/feeds/voltagelogger.voltage");
Adafruit_MQTT_Publish current = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "/feeds/voltagelogger.current");
Adafruit_MQTT_Publish trigger = Adafruit_MQTT_Publish(&mqtt, AIO_USERNAME "/feeds/voltagelogger.trigger");

// Setup a feed called 'onoff' for subscribing to changes.
Adafruit_MQTT_Subscribe onoffbutton = Adafruit_MQTT_Subscribe(&mqtt, AIO_USERNAME "/feeds/onoff");

/*************************** Sketch Code ************************************/

#define LEDPIN 13
#define TRIGGEROUT 11
#define TRIGGERIN 12
int battperc = 0;
int triggerCount = 0;

void setup() {

  pinMode(TRIGGEROUT, OUTPUT);
  digitalWrite(TRIGGEROUT, LOW);

  if (trackTriggerIn) {
    attachInterrupt(digitalPinToInterrupt(TRIGGERIN), receivedTriggerFunc, HIGH);
  }


  WiFi.setPins(WINC_CS, WINC_IRQ, WINC_RST, WINC_EN);

  //while (!Serial);
  Serial.begin(115200);
  delay(10);
  Serial.println("FeatherWing TFT Test!");

  tft.begin();
  tft.setRotation(2);

  tft.fillScreen(ILI9341_BLACK);

  tft.setCursor(0, 0);
  tft.setTextColor(ILI9341_WHITE);  tft.setTextSize(2);

  // Initialise the Client
  Serial.print(F("\nInit the WiFi module..."));
  // check for the presence of the breakout
  if (WiFi.status() == WL_NO_SHIELD) {
    Serial.println("WINC1500 not present");
    tft.println("WINC1500 not present");
    // don't continue:
  }
  Serial.println("ATWINC OK!");
  tft.println("WIFI OK");

  if (! ts.begin()) {
    Serial.println("STMPE not found!");
    while (1);
  }
  Serial.println("Waiting for touch sense");

  if (!SD.begin(SD_CS)) {
    Serial.println("Card failed, or not present");
    tft.println("SD Card failed, or not present");
    while (1);
  }
  Serial.println("card initialized.");
  tft.println("SD Card OK");

  if (SD.exists(datafile)) {
    SD.remove(datafile);
  }

  if (SD.exists(triggerfile)) {
    SD.remove(triggerfile);
  }

  File dataFile = SD.open(datafile, FILE_WRITE);
  String dataString = "date, bus voltage, shunt voltage, current (mA), load voltage, power";

  // if the file is available, write to it:
  if (dataFile) {
    dataFile.println(dataString);
    dataFile.close();
    // print to the serial port too:
  }

  File triggerFile = SD.open(triggerfile, FILE_WRITE);
  String headerString = "date, trigger";

  // if the file is available, write to it:
  if (triggerFile) {
    triggerFile.println(headerString);
    triggerFile.close();
    // print to the serial port too:
  }



  ina219.begin();
  // To use a slightly lower 32V, 1A range (higher precision on amps):
  //ina219.setCalibration_32V_1A();
  // Or to use a lower 16V, 400mA range (higher precision on volts and amps):
  ina219.setCalibration_16V_400mA();


  tft.fillScreen(ILI9341_BLACK);

  MQTT_connect();
  rtc.begin(); // initialize RTC
  rtc.setEpoch(WiFi.getTime());
  //getBattery();
  //drawButtons();

}

int period = 1000;
unsigned long time_print = 0;
unsigned long time_SD = 0;
unsigned long time_AIO = 0;
unsigned long time_TRIGGER = 0;

void loop() { //Main loop and write to SD


  //Print To Screen
  if (millis() > time_print + NORMAL_MILLISECS_BETWEEN_READING) {
    time_print = millis();
    getValue();
  }

  //SD
  if (millis() > time_SD + NORMAL_MILLISECS_BETWEEN_READING_SD) {
    getValue();
    time_SD = millis();
    if (sdActive) {
      recordValueSD();
    }
    Serial.println("SD");
  }

  //AIO
  if (millis() > time_AIO + NORMAL_MILLISECS_BETWEEN_READING_AIO) {
    getValue();
    time_AIO = millis();
    if (aioActive) {
      recordValueAIO();
    }
    Serial.println("AIO");
  }

  //Trigger
  if (millis() > time_TRIGGER + NORMAL_MILLISECS_BETWEEN_TRIGGER) {
    getValue();
    time_TRIGGER = millis();

    if (enableTrigger) {
      digitalWrite(TRIGGEROUT, HIGH);   // set trigger to high
      delay(200);
      shuntvoltage = ina219.getShuntVoltage_mV() + VOLTAGE_CALIBRATION;
      busvoltage = ina219.getBusVoltage_V() + VOLTAGE_CALIBRATION;
      current_mA = ina219.getCurrent_mA() + CURRENT_CALIBRATION;
      power_mW = ina219.getPower_mW();
      loadvoltage = busvoltage + (shuntvoltage / 1000);
      recordTrigger();
      recordValueAIO();
      recordValueSD();
      delay(200);
      digitalWrite(TRIGGEROUT, LOW);
    }
    Serial.println("TRIGGER");
  }

  if (recievedTrigger) {
    shuntvoltage = ina219.getShuntVoltage_mV() + VOLTAGE_CALIBRATION;
    busvoltage = ina219.getBusVoltage_V() + VOLTAGE_CALIBRATION;
    current_mA = ina219.getCurrent_mA() + CURRENT_CALIBRATION;
    power_mW = ina219.getPower_mW();
    loadvoltage = busvoltage + (shuntvoltage / 1000);
    recordValueAIO();
    recordValueSD();
    recordTrigger();
    recievedTrigger = false;
  }

}


void receivedTriggerFunc() {
  recievedTrigger = true;
}


void getValue() {
  shuntvoltage = ina219.getShuntVoltage_mV() + VOLTAGE_CALIBRATION;
  busvoltage = ina219.getBusVoltage_V() + VOLTAGE_CALIBRATION;
  current_mA = ina219.getCurrent_mA() + CURRENT_CALIBRATION;
  power_mW = ina219.getPower_mW();
  loadvoltage = busvoltage + (shuntvoltage / 1000);

  int fontHeight = 18;
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(0, 0);
  //void fillRect(uint16_t x0, uint16_t y0, uint16_t w, uint16_t h, uint16_t color);
  tft.fillRect(140, fontHeight * 0, 100, 15, ILI9341_BLACK);
  tft.print("Bus Volt:   "); tft.print(busvoltage); tft.println(" V");

  tft.setCursor(0, fontHeight * 1);
  tft.fillRect(140, fontHeight * 1, 100, 15, ILI9341_BLACK);
  tft.print("Shunt Volt: "); tft.print(shuntvoltage); tft.println(" mV");

  tft.setCursor(0, fontHeight * 2);
  tft.fillRect(140, fontHeight * 2, 100, 15, ILI9341_BLACK);
  tft.print("Load Vol:   "); tft.print(loadvoltage); tft.println(" V");

  tft.setCursor(0, fontHeight * 3);
  tft.fillRect(140, fontHeight * 3, 100, 15, ILI9341_BLACK);
  tft.print("Current:    "); tft.print(current_mA); tft.println(" mA");

  tft.setCursor(0, fontHeight * 4);
  tft.fillRect(140, fontHeight * 4, 100, 15, ILI9341_BLACK);
  tft.print("Power:      "); tft.print(power_mW); tft.println(" mW");
  //tft.println("");

  tft.setCursor(0, fontHeight * 5);
  tft.fillRect(140, fontHeight * 5, 100, 15, ILI9341_BLACK);
  tft.print("Triggers:      "); tft.println(triggerCount);
  //tft.println("");
}

void recordValueSD() {

  // make a string for assembling the data to log:
  String dataString = "";

  dataString = String(rtc.getEpoch()) + ",";
  dataString += String(busvoltage) + ",";
  dataString += String(shuntvoltage) + ",";
  dataString += String(current_mA) + ",";
  dataString += String(loadvoltage) + ",";
  dataString += String(power_mW) + ",";


  // open the file. note that only one file can be open at a time,
  // so you have to close this one before opening another.
  File dataFile = SD.open(datafile, FILE_WRITE);

  // if the file is available, write to it:
  if (dataFile) {
    dataFile.println(dataString);
    dataFile.close();
    // print to the serial port too:
  }
  // if the file isn't open, pop up an error:
  else {
    Serial.println("error opening datalog.txt");
  }

}

void recordValueAIO() {

  MQTT_connect();

  voltage.publish(busvoltage);
  current.publish(current_mA);

}

void recordTrigger() {

  if (aioActive) {
    MQTT_connect();
    trigger.publish("1");
  }

  if (sdActive) {

    String dataString =  String(rtc.getEpoch()) + ",1";
    File dataFile = SD.open(triggerfile, FILE_WRITE);

    // if the file is available, write to it:
    if (dataFile) {
      dataFile.println(dataString);
      dataFile.close();
      // print to the serial port too:
    }
  }

  triggerCount++;

}


// Function to connect and reconnect as necessary to the MQTT server.
// Should be called in the loop function and it will take care if connecting.
void MQTT_connect() {
  int8_t ret;

  // attempt to connect to Wifi network:
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print("Attempting to connect to SSID: ");
    Serial.println(ssid);
    // Connect to WPA/WPA2 network. Change this line if using open or WEP network:
    status = WiFi.begin(ssid, pass);

    // wait 10 seconds for connection:
    uint8_t timeout = 10;
    while (timeout && (WiFi.status() != WL_CONNECTED)) {
      timeout--;
      delay(1000);
    }
  }

  // Stop if already connected.
  if (mqtt.connected()) {
    return;
  }

  Serial.print("Connecting to MQTT... ");

  long minutes = 0; // Elapsed time / line number

  while ((ret = mqtt.connect()) != 0) { // connect will return 0 for connected
    Serial.println(mqtt.connectErrorString(ret));
    Serial.println("Retrying MQTT connection in 5 seconds...");
    mqtt.disconnect();
    delay(5000);  // wait 5 seconds
  }
  Serial.println("MQTT Connected!");
  Serial.println(WiFi.getTime());
}